; Compile stream of tokens to unoptimized intermediate ideal assembly code.

;
; DATA
;
C1MAXINS        = 0f8h                  ; maximum instructions per statement

C1EOL           = 0                     ; C1 intermediate instructions
C1NOP           = 1
C1ABS           = 2
C1NEG           = 3
C1NOT           = 4
C1IMUL          = 5
C1IDIV          = 6
C1MUL           = 7
C1DIV           = 8
C1MOD           = 9
C1ADD           = 10
C1SUB           = 11
C1SHL           = 12
C1SHR           = 13
C1SAR           = 14
C1ROL           = 15
C1ROR           = 16
C1AND           = 17
C1XOR           = 18
C1OR            = 19
C1MOV           = 20

align 4
oldstackptr     dd      ?

c1ins           db   22*C1MAXINS dup(?) ; C1 intermediate instructions

c1err0          db      'Empty statement.',CRLFT
c1err1          db      'Expecting end of statement '';''.',CRLFT
c1err2          db      'Syntax error.',CRLFT
c1err3          db      ''':'' equate override must be followed by valid equate.',CRLFT
c1err4          db      '''&'' offset override must be followed by valid label.',CRLFT
c1err5          db      'Missing right parentheses '')''.',CRLFT
c1err6          db      'Missing right bracket '']''.',CRLFT
c1err7          db      'Statement too complex (intermediate code too long).',CRLFT

;
; CODE
;

@newins         macro   ins
        mov al,&ins
        call c1_newins
endm

@newvar         macro
        mov ebx,ebp
        inc ebp
endm

;
c1_error:
        mov esp,oldstackptr
        call _io_error
        stc
        ret

;
c1_newins:
        mov c1ins[edi],al
        add edi,22
        cmp edi,C1MAXINS*22
        jb _ret
        mov edx,offset c1err7
        jmp c1_error

;
_c1:
        mov oldstackptr,esp
        mov edi,offset c1ins
        mov ecx,C1MAXINS*11
        xor eax,eax
        rep stosw
        mov ebp,8
        xor edi,edi
        mov ecx,1

        mov edx,offset c1err0
        cmp c0token[ecx*4],EOL
        je c1_error
        @newvar
        call c1_ass
        mov edx,offset c1err1
        cmp c0token[ecx*4],EOL
        jne c1_error
        clc
        ret

;
c1_ass:
        mov eax,c0token[ecx*4]
        cmp eax,ASCIIWORD
        je short c1_ass_mem
        cmp eax,REG
        je c1_ass_reg
        cmp eax,OP_MUL
        je c1_ass_ptr
        cmp eax,L_BRACKET
        jne c1_log
;-----------------------------------------------------------------------------
c1_ass_bracket:
        inc ecx
        call c1_ass
        mov edx,offset c1err6
        cmp c0token[ecx*4],R_BRACKET
        jne c1_error
        inc ecx
c1_ass_ptr_:
        mov c1ins[edi+12],1
        mov c1ins[edi+15],bl
        push ebx
        @newvar
        mov c1ins[edi+5],bl
        @newins C1MOV
        push ecx
        call c1_ass_
        pop edx
        pop eax
        cmp edx,ecx
        je c1_ass__
        mov c1ins[edi+2],1
        mov c1ins[edi+5],al
        mov c1ins[edi+15],bl
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_ass_mem:
        mov c1ins[edi+12],1
        mov c1ins[edi+16],cl
        mov c1ins[edi+5],bl
        inc ecx
        @newins C1MOV
        push ecx
        call c1_ass_
        pop edx
        cmp edx,ecx
        je short c1_ass__
        dec edx
        mov c1ins[edi+2],1
        mov c1ins[edi+6],dl
        mov c1ins[edi+15],bl
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_ass_reg:
        mov al,byte ptr c0tokenlptr[ecx*4]
        mov c1ins[edi+15],al
        mov c1ins[edi+5],bl
        inc ecx
        @newins C1MOV
        push ecx
        call c1_ass_
        pop edx
        cmp edx,ecx
        je short c1_ass__
        dec edx
        mov al,byte ptr c0tokenlptr[edx*4]
        mov c1ins[edi+5],al
        mov c1ins[edi+15],bl
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_ass_ptr:
        inc ecx
        push offset c1_ass_ptr_
        jmp c1_num
;-----------------------------------------------------------------------------
c1_ass__:
        push offset c1_log_
        push offset c1_shf_
        push offset c1_add_
        jmp c1_mul_
;-----------------------------------------------------------------------------
c1_ass_:
        mov eax,c0token[ecx*4]
        cmp eax,AS_SIMPLE
        je c1_ass_simple
        cmp eax,AS_MUL
        je c1_ass_mul
        cmp eax,AS_DIV
        je c1_ass_div
        cmp eax,AS_UMUL
        je c1_ass_umul
        cmp eax,AS_UDIV
        je c1_ass_udiv
        cmp eax,AS_MOD
        je c1_ass_mod
        cmp eax,AS_ADD
        je c1_ass_add
        cmp eax,AS_SUB
        je c1_ass_sub
        cmp eax,AS_AND
        je c1_ass_and
        cmp eax,AS_XOR
        je c1_ass_xor
        cmp eax,AS_OR
        je c1_ass_or
        cmp eax,AS_L_SHIFT
        je c1_ass_lshift
        cmp eax,AS_R_SHIFT
        je c1_ass_rshift
        cmp eax,AS_L_SSHIFT
        je c1_ass_lshift
        cmp eax,AS_R_SSHIFT
        je c1_ass_rsshift
        cmp eax,AS_L_ROT
        je c1_ass_lrot
        cmp eax,AS_R_ROT
        jne _ret
;-----------------------------------------------------------------------------
c1_ass_rrot:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1ROR
        ret
;-----------------------------------------------------------------------------
c1_ass_simple:
        inc ecx
        jmp c1_ass
;-----------------------------------------------------------------------------
c1_ass_mul:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1IMUL
        ret
;-----------------------------------------------------------------------------
c1_ass_div:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1IDIV
        ret
;-----------------------------------------------------------------------------
c1_ass_umul:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1MUL
        ret
;-----------------------------------------------------------------------------
c1_ass_udiv:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1DIV
        ret
;-----------------------------------------------------------------------------
c1_ass_mod:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1MOD
        ret
;-----------------------------------------------------------------------------
c1_ass_add:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1ADD
        ret
;-----------------------------------------------------------------------------
c1_ass_sub:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1SUB
        ret
;-----------------------------------------------------------------------------
c1_ass_and:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1AND
        ret
;-----------------------------------------------------------------------------
c1_ass_xor:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1XOR
        ret
;-----------------------------------------------------------------------------
c1_ass_or:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1OR
        ret
;-----------------------------------------------------------------------------
c1_ass_lshift:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1SHL
        ret
;-----------------------------------------------------------------------------
c1_ass_rshift:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1SHR
        ret
;-----------------------------------------------------------------------------
c1_ass_rsshift:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1SAR
        ret
;-----------------------------------------------------------------------------
c1_ass_lrot:
        inc ecx
        push ebx
        @newvar
        call c1_ass
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1ROL
        ret

;
c1_log:
        call c1_shf
c1_log_:
        mov eax,c0token[ecx*4]
        cmp eax,OP_AND
        je short c1_log_and
        cmp eax,OP_XOR
        je short c1_log_xor
        cmp eax,OP_OR
        jne _ret
;-----------------------------------------------------------------------------
c1_log_or:
        inc ecx
        push ebx
        @newvar
        call c1_shf
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1OR
        jmp c1_log_
;-----------------------------------------------------------------------------
c1_log_and:
        inc ecx
        push ebx
        @newvar
        call c1_shf
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1AND
        jmp c1_log_
;-----------------------------------------------------------------------------
c1_log_xor:
        inc ecx
        push ebx
        @newvar
        call c1_shf
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1XOR
        jmp c1_log_

;
c1_shf:
        call c1_add
c1_shf_:
        mov eax,c0token[ecx*4]
        cmp eax,OP_L_SHIFT
        je short c1_shf_lshift
        cmp eax,OP_R_SHIFT
        je short c1_shf_rshift
        cmp eax,OP_L_SSHIFT
        je short c1_shf_lshift
        cmp eax,OP_R_SSHIFT
        je short c1_shf_rsshift
        cmp eax,OP_L_ROT
        je c1_shf_lrot
        cmp eax,OP_R_ROT
        jne _ret
;-----------------------------------------------------------------------------
c1_shf_rrot:
        inc ecx
        push ebx
        @newvar
        call c1_add
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1ROR
        jmp c1_shf_
;-----------------------------------------------------------------------------
c1_shf_lshift:
        inc ecx
        push ebx
        @newvar
        call c1_add
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1SHL
        jmp c1_shf_
;-----------------------------------------------------------------------------
c1_shf_rshift:
        inc ecx
        push ebx
        @newvar
        call c1_add
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1SHR
        jmp c1_shf_
;-----------------------------------------------------------------------------
c1_shf_rsshift:
        inc ecx
        push ebx
        @newvar
        call c1_add
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1SAR
        jmp c1_shf_
;-----------------------------------------------------------------------------
c1_shf_lrot:
        inc ecx
        push ebx
        @newvar
        call c1_add
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1ROL
        jmp c1_shf_

;
c1_add:
        call c1_mul
c1_add_:
        mov eax,c0token[ecx*4]
        cmp eax,OP_ADD
        je short c1_add_add
        cmp eax,OP_SUB
        jne _ret
;-----------------------------------------------------------------------------
c1_add_sub:
        inc ecx
        push ebx
        @newvar
        call c1_mul
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1SUB
        jmp c1_add_
;-----------------------------------------------------------------------------
c1_add_add:
        inc ecx
        push ebx
        @newvar
        call c1_mul
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1ADD
        jmp c1_add_

;
c1_mul:
        call c1_una
c1_mul_:
        mov eax,c0token[ecx*4]
        cmp eax,OP_MUL
        je short c1_mul_mul
        cmp eax,OP_DIV
        je short c1_mul_div
        cmp eax,OP_UMUL
        je short c1_mul_umul
        cmp eax,OP_UDIV
        je c1_mul_udiv
        cmp eax,OP_MOD
        jne _ret
;-----------------------------------------------------------------------------
c1_mul_mod:
        inc ecx
        push ebx
        @newvar
        call c1_una
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1MOD
        jmp c1_mul_
;-----------------------------------------------------------------------------
c1_mul_mul:
        inc ecx
        push ebx
        @newvar
        call c1_una
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1IMUL
        jmp c1_mul_
;-----------------------------------------------------------------------------
c1_mul_div:
        inc ecx
        push ebx
        @newvar
        call c1_una
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1IDIV
        jmp c1_mul_
;-----------------------------------------------------------------------------
c1_mul_umul:
        inc ecx
        push ebx
        @newvar
        call c1_una
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1MUL
        jmp c1_mul_
;-----------------------------------------------------------------------------
c1_mul_udiv:
        inc ecx
        push ebx
        @newvar
        call c1_una
        mov c1ins[edi+15],bl
        pop ebx
        mov c1ins[edi+5],bl
        @newins C1DIV
        jmp c1_mul_

;
c1_una:
        mov eax,c0token[ecx*4]
        cmp eax,OP_NOT
        je short c1_una_not
        cmp eax,OP_SUB
        je short c1_una_neg
        cmp eax,OP_ADD
        jne short c1_num
;-----------------------------------------------------------------------------
c1_una_abs:
        inc ecx
        call c1_una
        mov c1ins[edi+5],bl
        @newins C1ABS
        ret
;-----------------------------------------------------------------------------
c1_una_not:
        inc ecx
        call c1_una
        mov c1ins[edi+5],bl
        @newins C1NOT
        ret
;-----------------------------------------------------------------------------
c1_una_neg:
        inc ecx
        call c1_una
        mov c1ins[edi+5],bl
        @newins C1NEG
        ret

;
c1_num:
        mov eax,c0token[ecx*4]
        cmp eax,ASCIIWORD
        je short c1_num_mem
        cmp eax,REG
        je short c1_num_reg
        cmp eax,IMMEDIATE
        je short c1_num_num
        cmp eax,EQUATE
        je c1_num_equate
        cmp eax,OP_AND
        je c1_num_off
        cmp eax,OP_MUL
        je c1_num_ptr
        cmp eax,L_PAREN
        je c1_num_paren
        cmp eax,L_BRACKET
        je c1_num_bracket
        mov edx,offset c1err2
        jmp c1_error
;-----------------------------------------------------------------------------
c1_num_mem:
        mov c1ins[edi+5],bl
        mov c1ins[edi+12],1
        mov c1ins[edi+16],cl
        inc ecx
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_num_reg:
        mov c1ins[edi+5],bl
        mov al,byte ptr c0tokenlptr[ecx*4]
        mov c1ins[edi+15],al
        inc ecx
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_num_num:
        mov c1ins[edi+5],bl
        mov eax,c0tokenlptr[ecx*4]
        mov dword ptr c1ins[edi+18],eax
        inc ecx
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_num_equate:
        inc ecx
        mov edx,offset c1err3
        cmp c0token[ecx*4],ASCIIWORD
        jne c1_error
        mov c1ins[edi+5],bl
        mov c1ins[edi+17],cl
        inc ecx
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_num_off:
        inc ecx
        mov edx,offset c1err4
        cmp c0token[ecx*4],ASCIIWORD
        jne c1_error
        mov c1ins[edi+5],bl
        mov c1ins[edi+16],cl
        inc ecx
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_num_ptr:
        inc ecx
        call c1_num
        mov c1ins[edi+5],bl
        mov c1ins[edi+12],1
        mov c1ins[edi+15],bl
        @newins C1MOV
        ret
;-----------------------------------------------------------------------------
c1_num_paren:
        inc ecx
        call c1_ass
        mov edx,offset c1err5
        cmp c0token[ecx*4],R_PAREN
        jne c1_error
        inc ecx
        ret
;-----------------------------------------------------------------------------
c1_num_bracket:
        inc ecx
        call c1_ass
        mov edx,offset c1err6
        cmp c0token[ecx*4],R_BRACKET
        jne c1_error
        mov c1ins[edi+5],bl
        mov c1ins[edi+12],1
        mov c1ins[edi+15],bl
        inc ecx
        @newins C1MOV
        ret

