; Convert ideal language to acutal 386 code and write to output data.

;
; DATA
;
C4MAXSTACK      = 88h

C4PUSH          = 21
C4POP           = 22
C4LEA           = 23
C4XCHG          = 24

align 4
c4optbl         dd      c4abs,c4una,c4una,c4imul,c4idiv,c4imul,c4div,c4mod
                dd      c4bin,c4bin,c4shf,c4shf,c4shf,c4shf,c4shf
                dd      c4bin,c4bin,c4bin,c4mov

c4reg           db      12*7 dup(?)
c4stack         db      12*C4MAXSTACK dup(?)

c4stacktop      db      ?

c4hexdigittbl   db      '0123456789abcdef'
c4regmask       db      0,2,4,8,10h,20h,40h,80h
c4regstr        db      'eax','ebx','ecx','edx'
                db      'esi','edi','ebp','esp'

c4strcomma      db      1,','
c4strplus       db      1,'+'
c4strscaleval2  db      2,'*2'          ;                                    |
c4strscaleval4  db      2,'*4'          ;                                    |
c4strscaleval8  db      2,'*8'          ;                                    |
c4strcrlf       db      2,0dh,0ah
c4stroffset     db      7,'offset '
c4straddxlds    db      14,'dword ptr ds:['     ;                            |
c4straddxlss    db      14,'dword ptr ss:['     ;                            |
c4straddxr      db      1,']'
c4straddxrcomma db      2,'],'
c4straddxrcrlf  db      3,']',0dh,0ah
c4strcommaaddxl db      2,',['
c4strcommaccrlf db      5,',cl',0dh,0ah
c4strlabel      db      13,'@@CAlabel0000'
c4strcoloncrlf  db      3,':',0dh,0ah
c4strcdqcrlf    db      13,'        cdq',0dh,0ah
c4strxoredxcrlf db      21,'        xor edx,edx',0dh,0ah
c4strjnsshort   db      18,'        jns short '

c4stropneg      db      4,'neg    '
c4stropnot      db      4,'not    '
c4stropimul     db      5,'imul   '
c4stropidiv     db      5,'idiv   '
c4stropmul      db      4,'mul    '     ; never used, 'imul' used instead
c4stropdiv      db      4,'div    '
c4stropmod      db      4,'div    '     ; never used, really 'mod'
c4stropadd      db      4,'add    '
c4stropsub      db      4,'sub    '
c4stropshl      db      4,'shl    '
c4stropshr      db      4,'shr    '
c4stropsar      db      4,'sar    '
c4stroprol      db      4,'rol    '
c4stropror      db      4,'ror    '
c4stropand      db      4,'and    '
c4stropxor      db      4,'xor    '
c4stropor       db      3,'or     '
c4stropmov      db      4,'mov    '
c4stroppush     db      5,'push   '
c4stroppop      db      4,'pop    '
c4stroplea      db      4,'lea    '
c4stropxchg     db      5,'xchg   '

;
; CODE
;

@putstr         macro   str
        mov esi,offset c4str&str
        call c4putstr
endm

@putnum         macro   num
        mov eax,&num
        call c4putnum
endm

@putop          macro   op
        mov al,&op
        call c4putop
endm

@putreg         macro   reg
        mov al,&reg
        call c4putreg
endm

;
c4outofmem:
        mov edx,offset ferr0
        jmp _io_fatalerror

;
; In: ESI -> counted string, Out: ESI - ?
c4putstr:
        push eax ecx edi
        mov edi,_ioocurptr
        movzx ecx,byte ptr [esi]
        inc esi
        lea eax,[edi+ecx]
        cmp eax,_iooendptr
        ja c4outofmem
        rep movsb
        mov _ioocurptr,edi
        pop edi ecx eax
        ret

;
; In: EAX - number, Out: EAX - ?
c4putnum:
        push ecx edx edi
        mov edi,_ioocurptr
        lea edx,[edi+10]
        cmp edx,_iooendptr
        ja c4outofmem
        mov byte ptr [edi],'0'
        inc edi
        mov edx,eax
        mov ecx,8
c4putnuml0:
        rol edx,4
        movzx eax,dl
        and al,0fh
        mov al,c4hexdigittbl[eax]
        stosb
        loop c4putnuml0
        mov byte ptr [edi],'h'
        inc edi
        mov _ioocurptr,edi
        pop edi edx ecx
        ret

;
; In: AL - opcode, Out: EAX - ?
c4putop:
        push ecx esi edi
        mov edi,_ioocurptr
        movzx eax,al
        lea esi,[eax*8+offset c4stropneg-8*3]
        movzx ecx,byte ptr [esi]
        inc esi
        lea eax,[edi+ecx+8]
        cmp eax,_iooendptr
        ja c4outofmem
        mov eax,20202020h
        stosd
        stosd
        rep movsb
        mov _ioocurptr,edi
        pop edi esi ecx
        ret

;
; In: AL - reg
c4putreg:
        push esi edi
        mov edi,_ioocurptr
        lea esi,[edi+3]
        cmp esi,_iooendptr
        ja c4outofmem
        movzx esi,al
        lea esi,[esi*2+esi+offset c4regstr-3]
        movsw
        movsb
        mov _ioocurptr,edi
        pop edi esi
        ret

;
; In: BX - reg (if 0, stack - BH=offset in stack (ESP+BH*4))
c4putloc:
        push eax
        or bl,bl
        jz short c4putlocf0
        @putreg bl
        pop eax
        ret
c4putlocf0:
        push esi
        @putstr addxlss
        @putreg 8
        or bh,bh
        jz short c4putlocf1
        @putstr plus
        movzx eax,bh
        shl eax,2
        call c4putnum
c4putlocf1:
        @putstr addxr
        pop esi eax
        ret

;
; In: EAX - label index, Out: EAX - ?
c4putlabel:
        push ecx esi edi
        mov edi,_ioocurptr
        mov esi,c0tokenlptr[eax*4]
        mov ecx,c0tokenllen[eax*4]
        lea eax,[edi+ecx]
        cmp eax,_iooendptr
        ja c4outofmem
        rep movsb
        mov _ioocurptr,edi
        pop edi esi ecx
        ret

;
; In: EDI -> expression, CL - base reg (if any), DL - index reg (if any)
c4putexp:
        push eax ebx esi
        xor bl,bl
        cmp byte ptr [edi+1],0
        je short c4putexpf5
        call c4putexpr0
        @putreg dl
        movzx esi,byte ptr [edi+2]
        or esi,esi
        jz short c4putexpf5
        lea esi,[esi*2+esi-3+offset c4strscaleval2]
        call c4putstr
c4putexpf5:
        cmp byte ptr [edi+3],0
        je short c4putexpf4
        call c4putexpr0
        @putreg cl
c4putexpf4:
        movzx eax,byte ptr [edi+4]
        or al,al
        jz short c4putexpf2
        call c4putexpr0
        @putstr offset
        call c4putlabel
c4putexpf2:
        movzx eax,byte ptr [edi+5]
        or al,al
        jz short c4putexpf3
        call c4putexpr0
        call c4putlabel
c4putexpf3:
        mov eax,[edi+6]
        or bl,bl
        jz short c4putexpf0
        or eax,eax
        jz short c4putexpf1
        @putstr plus
c4putexpf0:
        call c4putnum
c4putexpf1:
        pop esi ebx eax
        ret
;-----------------------------------------------------------------------------
c4putexpr0:
        bts ebx,0
        jnc _ret
        @putstr plus
        ret

;
c4l1r0:
        cmp al,[esi+1]
        je short c4l1r0f0
        cmp al,[esi+3]
        jne short c4l1r0f1
c4l1r0f0:
        mov byte ptr [esi+10],0
c4l1r0f1:
        add esi,12
        loop c4l1r0
        ret

;
_c4:
        mov c4stacktop,0
        shl c0regspreserve,1
        shl c3regsmodified,1

        mov cl,1
        mov ebx,offset c4reg
        xor eax,eax
c4l0:
        mov [ebx],eax
        mov [ebx+4],eax
        mov [ebx+8],eax
        mov [ebx+11],cl
        mov [ebx+3],cl
        mov byte ptr [ebx+10],1
        add ebx,12
        inc cl
        cmp cl,8
        jb c4l0

;-----------------------------------------------------------------------------
        mov ebp,offset c1ins
c4l1:
        movzx eax,byte ptr [ebp]
        cmp al,C1EOL
        je c4l1d
        cmp al,C1NOP
        je c4l1c

if DEBUG ne 0
        mov edi,_ioocurptr
        call pr0
        sub ebp,22
        mov _ioocurptr,edi
        movzx eax,byte ptr [ebp]
endif
        call c4optbl[eax*4-C1ABS*4]

        cmp byte ptr [ebp+2],0
        jne short c4l1f1
        mov al,[ebp+5]
        cmp al,8
        jb short c4l1f1
        cmp byte ptr [ebp],C1MOV
        je short c4l1f2
        call c4ft
        movzx ebx,bl
        lea ebx,[ebx*2+ebx-3]
        mov c4reg[ebx*4+10],0
c4l1f2:
        movzx ecx,c4stacktop
        jecxz short c4l1f0
        mov ebx,ecx
        mov esi,offset c4stack+11
c4l1l0:
        cmp al,[esi]
        jne short c4l1l0c
        mov byte ptr [esi],0
c4l1l0c:
        add esi,12
        dec ebx
        jnz c4l1l0
        mov esi,offset c4stack
        call c4l1r0
c4l1f0:
        mov ecx,7
        mov esi,offset c4reg
        call c4l1r0
c4l1f1:
if DEBUG ne 0
        call p2
endif

c4l1c:
        add ebp,22
        jmp c4l1
c4l1d:

;-----------------------------------------------------------------------------
        mov esi,offset c4reg
        mov ax,201h
c4l2:
        test c0regspreserve,ah
        jnz short c4l2f0
        test c3regsmodified,ah
        jz c4l2c
        or al,0f8h
c4l2f0:
        call c4ft
        and al,7
        or bl,bl
        jz short c4l2f1
        cmp al,bl
        je c4l2c
        call c4rx
        jmp c4l2c
c4l2f1:
        movzx ecx,byte ptr [esi+11]
        cmp cl,al
        jbe short c4l2f2
        cmp cl,8
        jb short c4l2f3
        cmp cl,0f8h
        jbe short c4l2f2
        lea edi,[ecx*2+ecx-3*0f9h]
        lea edi,[edi*4+offset c4reg]
        cmp byte ptr [edi+11],cl
        je short c4l2f2
        xchg ebx,ecx
        and bl,7
        call c4rx
        mov ebx,ecx
        jmp c4l2f1
c4l2f3:
        bt dword ptr c0regspreserve,ecx
        jnc short c4l2f2
        lea edi,[ecx*2+ecx-3]
        lea edi,[edi*4+offset c4reg]
        cmp byte ptr [edi+11],cl
        je short c4l2f2
        xchg ebx,ecx
        call c4rx
        mov ebx,ecx
        jmp c4l2f1
c4l2f2:
        push offset c4l2c
        or bh,bh
        jz c4po
        jmp c4rl
c4l2c:
        add esi,12
        inc al
        shl ah,1
        jnz c4l2

;-----------------------------------------------------------------------------
        movzx ebx,c4stacktop
        shl ebx,2
        jz short c4f0
        @putop C1ADD
        @putreg 8
        @putstr comma
        @putnum ebx
        @putstr crlf
c4f0:
if DEBUG ne 0
        mov c4stacktop,0
        call p2
endif
        ret

;
c4abs:
        mov al,[ebp+5]
        call c4ft
        or bl,bl
        jnz short c4absf0
        call c4arbsr
        call c4rl
        mov bl,al
c4absf0:
        @putop C1OR
        mov al,bl
        call c4putreg
        @putstr comma
        call c4putreg
        @putstr crlf
        @putstr jnsshort
        @putstr label
        @putstr crlf
        @putop C1NEG
        @putreg bl
        @putstr crlf
        @putstr label
        @putstr coloncrlf
        mov ax,word ptr c4strlabel[12]
        mov bx,word ptr c4strlabel[10]
        inc ah
        cmp ah,'9'
        jbe short c4absf1
        mov ah,'0'
        inc al
        cmp al,'9'
        jbe short c4absf1
        mov al,'0'
        inc bh
        cmp bh,'9'
        jbe short c4absf1
        mov bh,'0'
        inc bl
        cmp bl,'9'
        jbe short c4absf1
        mov bl,'0'
c4absf1:
        mov word ptr c4strlabel[12],ax
        mov word ptr c4strlabel[10],bx
        ret
;
c4una:
        mov al,[ebp+5]
        call c4ft
        or bl,bl
        jnz short c4unaf0
        call c4arbsr
        call c4rl
        mov bl,al
c4unaf0:
        @putop [ebp]
        @putreg bl
        @putstr crlf
        ret
;
c4imul:
        call c4rsex
        cmp bl,0
        jg short c4imulf0
        cmp al,1
        je c4imulf1
        ja short c4imulf2
        mov ebx,ecx
        jmp short c4imulf0
c4imulf2:
        cmp bl,0f8h
        jne short c4imulf0
        mov edi,ebx
        mov esi,eax
        movzx eax,byte ptr [ebp+5]
        call c4ft
        or bl,bl
        jz short c4imulf4
        movzx ebx,bl
        mov ah,c4regmask[ebx]
c4imulf4:
        xchg esi,eax
        or eax,esi
        mov ebx,edi
        call c4rslr
        cmp byte ptr [ebp+12],0
        je short c4imulf5
        call c4imulr0
        @putstr addxlds
        lea edi,[ebp+12]
        call c4putexp
        @putstr addxrcrlf
        ret
c4imulf5:
        mov eax,esi
        lea edi,[ebp+12]
        call c4ce
        call c4rslv
        mov bl,al
c4imulf0:
        movzx eax,bl
        mov ah,c4regmask[eax]
        call c4imulr0
        call c4putloc
        @putstr crlf
        ret
c4imulf1:
        @putop C1IMUL
        mov al,[ebp+5]
        call c4ft
        or bl,bl
        jnz short c4imulf3
        call c4arbsr
        call c4putreg
        @putstr comma
        movzx edi,al
        lea edi,[edi*2+edi-3]
        movzx eax,c4stacktop
        sub al,bh
        lea esi,[eax*2+eax-3]
        mov al,c4stack[esi*4+11]
        mov c4reg[edi*4+11],al
c4imulf3:
        call c4putloc
        @putstr comma
        lea edi,[ebp+12]
        call c4putexp
        @putstr crlf
        ret
;-----------------------------------------------------------------------------
c4imulr0:
        mov edi,ebx
        mov al,[ebp+5]
        call c4ft
        or bl,bl
        jnz short c4imulr0f0
        call c4arbsr
        call c4rl
        mov bl,al
c4imulr0f0:
        @putop C1IMUL
        @putreg bl
        @putstr comma
        mov ebx,edi
        ret
;
c4idiv:
        push offset c4strcdqcrlf
        call c4idm
        mov al,[ebp+5]
        mov c4reg[11],al
        mov word ptr c4reg[36+10],0
        ret
;-----------------------------------------------------------------------------
c4div:
        push offset c4strxoredxcrlf
        call c4idm
        mov al,[ebp+5]
        mov c4reg[11],al
        mov word ptr c4reg[36+10],0
        ret
;-----------------------------------------------------------------------------
c4mod:
        push offset c4strxoredxcrlf
        call c4idm
        mov al,[ebp+5]
        mov c4reg[36+11],al
        mov word ptr c4reg[10],0
        ret
;-----------------------------------------------------------------------------
c4idm:
        call c4rsex
        cmp bl,0
        jg short c4idmf0
        cmp al,1
        ja short c4idmf1
        je short c4idmf2
        mov ebx,ecx
        jmp short c4idmf0
c4idmf1:
        cmp bl,0f8h
        jne short c4idmf0
        call c4idmr1
        call c4rslr
        cmp byte ptr [ebp+12],0
        je short c4idmf3
        or eax,esi
        cmp byte ptr [ebp+13],0
        je short c4idmf4
        xchg ebx,ecx
        call c4idmr2
        xchg ebx,ecx
c4idmf4:
        cmp byte ptr [ebp+15],0
        je short c4idmf5
        xchg ebx,edx
        call c4idmr2
        xchg ebx,edx
c4idmf5:
        and ah,0ech
        call c4idmr0
        @putstr addxlds
        lea edi,[ebp+12]
        call c4putexp
        @putstr addxrcrlf
        ret 4
c4idmf2:
        call c4idmr1
c4idmf3:
        mov eax,esi
        lea edi,[ebp+12]
        call c4ce
        call c4rslv
        mov bl,al
c4idmf0:
        movzx eax,bl
        mov ah,c4regmask[eax]
        call c4idmr0
        call c4putloc
        @putstr crlf
        ret 4
;-----------------------------------------------------------------------------
c4idmr0:
        push ebx
        mov [esp+2],ah
        mov al,[ebp+5]
        call c4ft
        cmp bl,1
        je short c4idmr0f0
        or bl,bl
        jnz short c4idmr0f1
        test ah,2
        jnz short c4idmr0f2
        mov ah,0fch
c4idmr0f2:
        call c4arbsr
        call c4rl
        movzx ebx,al
        mov ah,c4regmask[ebx]
        or ah,[esp+2]
        cmp bl,1
        je short c4idmr0f0
c4idmr0f1:
        mov al,1
        cmp [esp],al
        jne short c4idmr0f3
        movzx edi,bl
        or ah,c4reg[edi]
        mov [esp],bl
c4idmr0f3:
        call c4rx
c4idmr0f0:
        or ah,2
        mov bl,[esp]
        cmp bl,4
        jne short c4idmr0f4
        call c4arb
        mov [esp],al
        pushfd
        call c4rx
        popfd
        jnc short c4idmr0f5
        mov al,bl
        jmp short c4idmr0f6
c4idmr0f4:
        mov al,4
        call c4fur
        jc short c4idmr0f5
c4idmr0f6:
        call c4sr
c4idmr0f5:
        mov esi,[esp+12]
        call c4putstr
        @putop [ebp]
        pop ebx
        ret
;-----------------------------------------------------------------------------
c4idmr1:
        mov edi,ebx
        mov esi,eax
        movzx eax,byte ptr [ebp+5]
        call c4ft
        or bl,bl
        jz short c4idmr1f0
        movzx ebx,bl
        mov ah,c4regmask[ebx]
c4idmr1f0:
        or ah,12h
        xchg esi,eax
        or eax,esi
        mov ebx,edi
        ret
;-----------------------------------------------------------------------------
c4idmr2:
        cmp bl,4
        jne _ret
        call c4arb
        call c4rx
        movzx ebx,al
        or ah,c4regmask[ebx]
        ret
;
c4bin:
        call c4rsex
        cmp bl,0
        jg short c4binf0
        cmp al,1
        je c4binf1
        ja short c4binf2
        mov ebx,ecx
        jmp short c4binf0
c4binf2:
        cmp bl,0f8h
        jne short c4binf0
        mov edi,ebx
        mov esi,eax
        movzx eax,byte ptr [ebp+5]
        call c4ft
        or bl,bl
        jz short c4binf3
        movzx ebx,bl
        mov ah,c4regmask[ebx]
c4binf3:
        xchg esi,eax
        or eax,esi
        mov ebx,edi
        call c4rslr
        cmp byte ptr [ebp+12],0
        je short c4binf4
        call c4binr0
        @putstr addxlds
        lea edi,[ebp+12]
        call c4putexp
        @putstr addxrcrlf
        ret
c4binf4:
        mov eax,esi
        lea edi,[ebp+12]
        call c4ce
        call c4rslv
        mov bl,al
c4binf0:
        movzx eax,bl
        mov ah,c4regmask[eax]
        call c4binr0
        call c4putloc
        @putstr crlf
        ret
c4binf1:
        call c4binr0
        lea edi,[ebp+12]
        call c4putexp
        @putstr crlf
        ret
;-----------------------------------------------------------------------------
c4binr0:
        mov edi,ebx
        mov al,[ebp+5]
        call c4ft
        or bl,bl
        jnz short c4binr0f0
        call c4arbsr
        call c4rl
        mov bl,al
c4binr0f0:
        @putop [ebp]
        @putreg bl
        @putstr comma
        mov ebx,edi
        ret
;
c4shf:
        call c4rsex
        cmp bl,0
        jg c4shff0
        cmp al,1
        je short c4shff1
        ja short c4shff2
        or cl,cl
        jz short c4shff4
        mov bl,cl
        jmp c4shff0
c4shff4:
        mov al,[ebp+5]
        call c4ft
        mov ah,0f6h
        cmp bl,3
        jne short c4shff5
        mov ah,8
c4shff5:
        call c4arbsr
        mov bx,cx
        call c4rl
        mov bl,al
        jmp short c4shff0
c4shff1:
        cmp byte ptr [ebp+16],0
        jne short c4shff2
        call c4shfr0
        @putop [ebp]
        @putreg bl
        @putstr comma
        lea edi,[ebp+12]
        call c4putexp
        @putstr crlf
        ret
c4shff2:
        mov edi,ebx
        mov al,[ebp+5]
        call c4ft
        or bl,bl
        jz short c4shff6
        movzx ebx,bl
        or ah,c4regmask[ebx]
c4shff6:
        call c4rslr
        mov ah,0f6h
        cmp bl,3
        jne short c4shff7
        mov ah,8
c4shff7:
        mov ebx,edi
        lea edi,[ebp+12]
        call c4ce
        call c4rslv
        mov bl,al
c4shff0:
        cmp bl,3
        je short c4shff3
        mov al,3
        call c4rx
c4shff3:
        mov ah,8
        call c4shfr0
        @putop [ebp]
        @putreg bl
        @putstr commaccrlf
        ret
;-----------------------------------------------------------------------------
c4shfr0:
        mov al,[ebp+5]
        call c4ft
        or bl,bl
        jnz _ret
        call c4arbsr
        call c4rl
        mov bl,al
        ret
;
c4mov:
        call c4rsex
        cmp byte ptr [ebp+2],0
        jne c4movf0
        test ah,1
        jz short c4movf1
        cmp bl,0
        jg short c4movf1
        jl short c4movf3
        call c4movr0
        jmp short c4movf4
c4movf3:
        call c4rslr
c4movf1:
        call c4movr0
        lea edi,[ebp+12]
        call c4ce
        call c4rslv
        movzx ebx,al
        lea edi,[ebx*2+ebx-3]
        mov al,c4reg[edi*4+11]
        or al,al
        jz short c4movf2
        call c4fut
        jc short c4movf2
c4movf4:
        call c4arbsr
        call c4rl
        movzx ebx,al
        lea edi,[ebx*2+ebx-3]
c4movf2:
        mov eax,esi
        mov c4reg[edi*4+11],al
        ret
c4movf0:
        cmp bl,0
        jg short c4movf5
        cmp al,1
        je short c4movf6
        ja short c4movf7
        or cl,cl
        jz short c4movf7
        mov bl,cl
        jmp short c4movf5
c4movf7:
        mov edi,ebx
        mov esi,eax
        movzx eax,byte ptr [ebp+3]
        or al,al
        jz short c4movf8
        call c4ft
        or bl,bl
        jz short c4movf8
        movzx ebx,bl
        or ah,c4regmask[ebx]
c4movf8:
	mov al,[ebp+5]
        or al,al
	jz short c4movf9
        call c4ft
        or bl,bl
	jz short c4movf9
        movzx ebx,bl
        or ah,c4regmask[ebx]
c4movf9:
	xchg esi,eax
        or eax,esi
        mov ebx,edi
        call c4rslr
        mov eax,esi
        lea edi,[ebp+12]
        call c4ce
        call c4rslv
        mov bl,al
c4movf5:
        movzx ebx,bl
        mov ah,c4regmask[ebx]
        call c4movr1
        @putreg bl
        @putstr crlf
        ret
c4movf6:
        call c4movr1
        lea edi,[ebp+12]
        call c4putexp
        @putstr crlf
        ret
;-----------------------------------------------------------------------------
c4movr0:
        mov edi,ebx
        movzx eax,byte ptr [ebp+5]
        cmp al,7
        ja short c4movr0f0
        add al,0f8h
c4movr0f0:
        mov esi,eax
        call c4ft
        jc short c4movr0f1
        or bl,bl
        jz short c4movr0f1
        movzx ebx,bl
        mov al,c4regmask[ebx]
        not al
        and ah,al
        lea ebx,[ebx*2+ebx-3]
        mov byte ptr c4reg[ebx*4+11],0
c4movr0f1:
        mov ebx,edi
        ret
;-----------------------------------------------------------------------------
c4movr1:
        mov edi,ebx
        mov dl,0f8h
        mov al,[ebp+3]
        or al,al
        jz short c4movr1f0
        call c4ft
        mov edx,ebx
        or bl,bl
        jz short c4movr1f0
        movzx ebx,bl
        or ah,c4regmask[ebx]
c4movr1f0:
        mov al,[ebp+5]
        or al,al
        jz short c4movr1f1
        call c4ft
        mov ecx,ebx
        or bl,bl
        jnz short c4movr1f2
        call c4arbsr
        call c4rl
        mov bl,al
c4movr1f2:
        cmp dx,cx
        jne short c4movr1f4
        movzx edx,bl
c4movr1f4:
        movzx ecx,bl
        or ah,c4regmask[ecx]
c4movr1f1:
        or dl,dl
        jnz short c4movr1f3
        mov ebx,edx
        call c4arbsr
        call c4rl
        movzx edx,al
        or ah,c4regmask[edx]
c4movr1f3:
        mov ebx,edi
        @putop C1MOV
        @putstr addxlds
        lea edi,[ebp+2]
        call c4putexp
        @putstr addxrcomma
        ret

;
c4arsr:
        call c4ar
        jnc _ret
        jmp c4sr
;-----------------------------------------------------------------------------
c4arbsr:
        call c4arb
        jnc _ret
        jmp c4sr

;
; Classify expression
; In:
;   EDI -> expression
; Out:
;   AL - 0=temporary, 1=immediate, 2=expression, 3=immediate memory
;
c4ce:
        mov al,2
        cmp byte ptr [edi],0
        jne short c4cef1
        cmp word ptr [edi+1],0
        jne _ret
        cmp byte ptr [edi+3],0
        je short c4cef0
        cmp word ptr [edi+4],0
        jne _ret
        cmp dword ptr [edi+6],0
        jne _ret
        xor al,al
        ret
c4cef0:
        mov al,1
        ret
c4cef1:
        cmp word ptr [edi+1],0
        jne _ret
        cmp byte ptr [edi+3],0
        jne _ret
        mov al,3
        ret

;
; Find temporary in registers or stack
; In:
;   AL - temporary to find
; Out:
;   CF=1 - not found
;   CF=0 - found
;   BX - location (BL=reg or 0 means stack loc BH (ESP+BH*4)) (if any)
;
c4ft:
        push esi
        mov esi,offset c4reg+12*6+11
        mov bl,7
c4ftl0:
        cmp al,[esi]
        je short c4ftd
        sub esi,12
        dec bl
        jnz c4ftl0
        movzx esi,c4stacktop
        lea esi,[esi*2+esi]
        shl esi,2
        mov bh,-1
c4ftl1:
        inc bh
        sub esi,12
        jc short c4ftd
        cmp al,c4stack[esi+11]
        jne c4ftl1
c4ftd:
        pop esi
        ret

;
; Find expression in registers or stack
; In:
;   EDI -> expression to find
; Out:
;   CF=1 - not found
;   CF=0 - found
;   BX - location (BL=reg or 0 means stack loc BH (ESP+BH*4)) (if any)
;
c4fe:
        push eax ecx edx esi
        mov esi,offset c4reg+12*6
        mov eax,[edi]
        mov ecx,[edi+4]
        mov dx,[edi+8]
        mov bl,7
c4fel0:
        cmp byte ptr [esi+10],0
        je short c4fel0c
        cmp eax,[esi]
        jne short c4fel0c
        cmp ecx,[esi+4]
        jne short c4fel0c
        cmp dx,[esi+8]
        je short c4fed
c4fel0c:
        sub esi,12
        dec bl
        jnz c4fel0
        movzx esi,c4stacktop
        lea esi,[esi*2+esi]
        shl esi,2
        mov bh,-1
c4fel1:
        inc bh
        sub esi,12
        jc short c4fed
        cmp c4stack[esi+10],0
        je c4fel1
        cmp eax,dword ptr c4stack[esi]
        jne c4fel1
        cmp ecx,dword ptr c4stack[esi+4]
        jne c4fel1
        cmp dx,word ptr c4stack[esi+8]
        jne c4fel1
c4fed:
        pop esi edx ecx eax
        ret

;
; Push reg to top of stack
; In:
;   AL - reg
;
c4pu:
        push eax esi edi
        movzx edi,al
        @putop C4PUSH
        mov eax,edi
        call c4putreg
        @putstr crlf
        lea esi,[edi*2+edi-3]
        lea esi,[esi*4+offset c4reg]
        movzx eax,c4stacktop
        lea edi,[eax*2+eax]
        lea edi,[edi*4+offset c4stack]
        inc eax
        mov c4stacktop,al
        movsd
        movsd
        movsd
        pop edi esi eax
        ret

;
; Pop reg from top of stack
; In:
;   AL - reg
;
c4po:
        push eax esi edi
        movzx edi,al
        @putop C4POP
        mov eax,edi
        call c4putreg
        @putstr crlf
        lea edi,[edi*2+edi-3]
        lea edi,[edi*4+offset c4reg]
        movzx eax,c4stacktop
        dec eax
        mov c4stacktop,al
        lea esi,[eax*2+eax]
        lea esi,[esi*4+offset c4stack]
        movsd
        movsd
        movsd
        pop edi esi eax
        ret

;
; Store register to location
; In:
;   AL - reg
;   BX - location (BL=reg or 0 means stack loc BH (ESP+BH*4))
;
c4rs:
        push eax esi edi
        @putop C1MOV
        call c4putloc
        @putstr comma
        @putreg [esp+8]
        @putstr crlf
        movzx esi,byte ptr [esp+8]
        lea esi,[esi*2+esi-3]
        lea esi,[esi*4+offset c4reg]
        or bl,bl
        jz short c4rsf0
        movzx edi,bl
        lea edi,[edi*2+edi-3]
        lea edi,[edi*4+offset c4reg]
        jmp short c4rsf1
c4rsf0:
        movzx eax,c4stacktop
        sub al,bh
        lea edi,[eax*2+eax-3]
        lea edi,[edi*4+offset c4stack]
c4rsf1:
        movsd
        movsd
        movsd
        pop edi esi eax
        ret

;
; Load register from location
; In:
;   AL - reg
;   BX - location (BL=reg or 0 means stack loc BH (ESP+BH*4))
;
c4rl:
        push eax esi edi
        @putop C1MOV
        @putreg [esp+8]
        @putstr comma
        call c4putloc
        @putstr crlf
        movzx edi,byte ptr [esp+8]
        lea edi,[edi*2+edi-3]
        lea edi,[edi*4+offset c4reg]
        or bl,bl
        jz short c4rlf0
        movzx esi,bl
        lea esi,[esi*2+esi-3]
        lea esi,[esi*4+offset c4reg]
        jmp short c4rlf1
c4rlf0:
        movzx eax,c4stacktop
        sub al,bh
        lea esi,[eax*2+eax-3]
        lea esi,[esi*4+offset c4stack]
c4rlf1:
        movsd
        movsd
        movsd
        pop edi esi eax
        ret

;
; Exchange register with location
; In:
;   AL - reg
;   BX - location (BL=reg or 0 means stack loc BH (ESP+BH*4))
;
c4rx:
        push eax esi edi
        @putop C4XCHG
        @putreg [esp+8]
        @putstr comma
        call c4putloc
        @putstr crlf
        movzx edi,byte ptr [esp+8]
        lea edi,[edi*2+edi-3]
        lea edi,[edi*4+offset c4reg]
        or bl,bl
        jz short c4rxf0
        movzx esi,bl
        lea esi,[esi*2+esi-3]
        lea esi,[esi*4+offset c4reg]
        jmp short c4rxf1
c4rxf0:
        movzx eax,c4stacktop
        sub al,bh
        lea esi,[eax*2+eax-3]
        lea esi,[esi*4+offset c4stack]
c4rxf1:
        mov eax,[esi]
        xchg eax,[edi]
        mov [esi],eax
        mov eax,[esi+4]
        xchg eax,[edi+4]
        mov [esi+4],eax
        mov eax,[esi+8]
        xchg eax,[edi+8]
        mov [esi+8],eax
        pop edi esi eax
        ret

;
; Find nearest use of temporary in code that follows
; In:
;   AL - temporary to find
;   EBP -> current instruction
; Out:
;   CF=1 - use not found
;   CF=0 - use found
;     CL - distance to use (in instructions (0ffh means no use, but preserve))
;
c4fut:
        push ebx ebp
        cmp al,0f8h
        ja short c4futf0
        xor bl,bl
c4futl0:
        add ebp,22
        inc bl
        mov bh,[ebp]
        cmp bh,C1NOP
        jc short c4futd
        je c4futl0
        cmp bh,C1MOV
        jne short c4futl0f0
        cmp byte ptr [ebp+2],0
        je short c4futl0f1
        cmp al,[ebp+3]
        je short c4futd3
c4futl0f0:
        cmp al,[ebp+5]
        je short c4futd3
        cmp bh,C1NOT
        jbe c4futl0
c4futl0f1:
        cmp al,[ebp+13]
        je short c4futd3
        cmp al,[ebp+15]
        jne c4futl0
c4futd3:
        mov cl,bl
        clc
c4futd2:
        pop ebp ebx
        ret
c4futd:
        cmp al,8
        cmc
        jc c4futd2
        movzx ebx,al
        bt dword ptr c0regspreserve,ebx
        cmc
        jc c4futd2
c4futf0:
        mov cl,-1
        clc
        jmp c4futd2

;
; Find nearest use of expression in code that follows
; In:
;   EDI -> expression to find
;   EBP -> current instruction
; Out:
;   CF=1 - use not found
;   CF=0 - use found
;     CL - distance to use (in instructions (0ffh means no use, but preserve))
;
c4fue:
        push eax ebx edx esi edi ebp
        mov al,[edi+1]
        cmp al,8
        jae short c4fuef0
        mov al,0f8h
c4fuef0:
        mov ah,[edi+3]
        cmp ah,8
        jae short c4fuef1
        mov ah,0f8h
c4fuef1:
        mov edx,[edi]
        mov esi,[edi+4]
        mov di,[edi+8]
        xor bl,bl
c4fuel0:
        add ebp,22
        inc bl
        mov bh,[ebp]
        cmp bh,C1NOP
        jc short c4fuel0d
        je c4fuel0
        cmp bh,C1NOT
        jbe short c4fuel0c
        cmp edx,[ebp+12]
        jne short c4fuel0c
        cmp esi,[ebp+16]
        jne short c4fuel0c
        cmp di,[ebp+20]
        je short c4fuel0d2
c4fuel0c:
        cmp byte ptr [ebp+2],0
        jne c4fuel0
        cmp al,[ebp+5]
        stc
        je short c4fuel0d
        cmp ah,[ebp+5]
        stc
        je short c4fuel0d
        jmp c4fuel0
c4fuel0d2:
        mov cl,bl
        clc
c4fuel0d:
        pop ebp edi esi edx ebx eax
        ret

;
; Find nearest use of register in code that follows
; In:
;   AL - register to find
;   EBP -> current instruction
; Out:
;   CF=1 - use not found
;   CF=0 - use found
;     CL - distance to use (in instructions (0ffh means no use, but preserve))
;
c4fur:
        push eax edi
        movzx edi,al
        lea edi,[edi*2+edi-3]
        lea edi,[edi*4+offset c4reg]
        movzx eax,byte ptr [edi+11]
        or al,al
        jz short c4furf0
        call c4fut
        jc short c4furf0
        mov ah,cl
c4furf0:
        cmp byte ptr [edi+10],0
        je short c4furf1
        call c4ce
        cmp al,2
        jne short c4furf1
        call c4fue
        jc short c4furf1
        or ah,ah
        jz short c4furd
        cmp ah,cl
        jae short c4furd
        mov cl,ah
c4furf1:
        cmp ah,1
c4furd:
        pop edi eax
        ret

;
; Allocate register that is used furthest away
; In:
;   AH - valid registers (bits 1-7: 0 means valid) (must be at least one)
;   EBP -> current instruction
; Out:
;   CF=1 - register needs to be saved
;   CF=0 - register does not need to be saved
;   AL - register
;
c4ar:
        push ecx edx
        xor edx,edx
        mov ch,2
        mov al,1
c4arl0:
        test ah,ch
        jnz short c4arl0c
        call c4fur
        cmc
        jnc short c4ard
        cmp cl,dh
        jbe short c4arl0c
        mov dh,cl
        mov dl,al
c4arl0c:
        inc al
        shl ch,1
        jnz c4arl0
        mov al,dl
        stc
c4ard:
        pop edx ecx
        ret

;
; Allocate register that is used furthest away with bias
; In:
;   AH - valid registers (bits 1-7: 0 means valid) (must be at least one)
;   EBP -> current instruction
; Out:
;   CF=1 - register needs to be saved
;   CF=0 - register does not need to be saved
;   AL - register
;
c4arb:
        push ecx ebx edx
        xor bh,bh
        cmp byte ptr [ebp+2],0
        jne short c4arbf0
        mov bh,[ebp+5]
        cmp bh,8
        jb short c4arbf0
        movzx ebx,bh
        mov bh,c5regbias[ebx-8]
c4arbf0:
        xor edx,edx
        mov ch,2
        xor bl,bl
        mov al,1
c4arbl0:
        test ah,ch
        jnz short c4arbl0c
        call c4fur
        cmc
        jc short c4arbl0f1
        cmp al,bh
        je short c4arbl0f2
        or bl,bl
        jnz short c4arbl0c
c4arbl0f2:
        mov bl,al
        jmp short c4arbl0c
c4arbl0f1:
        cmp cl,dh
        jb short c4arbl0c
        ja short c4arbl0f0
        cmp al,bh
        jne short c4arbl0c
c4arbl0f0:
        mov dh,cl
        mov dl,al
c4arbl0c:
        inc al
        shl ch,1
        jnz c4arbl0
        mov al,bl
        or al,al
        jnz short c4arbd
        mov al,dl
        stc
c4arbd:
        pop edx ebx ecx
        ret

;
; Store register somewhere and null it out
; In:
;   AL - register to store
;   AH - valid registers (bits 1-7: 0 means valid)
;
c4sr:
        push eax ebx ecx esi edi
        movzx ebx,al
        lea esi,[ebx*2+ebx-3]
        lea esi,[esi*4+offset c4reg]
        or ah,c4regmask[ebx]
        cmp ah,0feh
        je short c4srf0
        mov bl,al
        call c4ar
        xchg al,bl
        jnc c4srf1
c4srf0:
        movzx edi,c4stacktop
        lea edi,[edi*2+edi]
        lea edi,[edi*4+c4stack]
        xor ah,ah
        mov bl,0ffh
c4srl0:
        sub edi,12
        cmp edi,offset c4stack
        jb c4srl0d
        mov cx,[edi+10]
        cmp ch,[esi+11]
        jne short c4srl0f1
        or ch,ch
        jnz c4srf2
        or cl,cl
        jz short c4srl0f0
        mov ecx,[edi]
        cmp ecx,[esi]
        jne short c4srl0c
        mov ecx,[edi+4]
        cmp ecx,[esi+4]
        jne short c4srl0c
        mov cx,[edi+8]
        cmp cx,[esi+8]
        jne short c4srl0c
        jmp short c4srf2
c4srl0f1:
        or ch,ch
        jnz short c4srl0f2
        or cl,cl
        jz short c4srl0f0
        cmp byte ptr [esi+10],0
        je short c4srl0c
        mov ecx,[edi]
        cmp ecx,[esi]
        jne short c4srl0c
        mov ecx,[edi+4]
        cmp ecx,[esi+4]
        jne short c4srl0c
        mov cx,[edi+8]
        cmp cx,[esi+8]
        jne short c4srl0c
        jmp short c4srf2
c4srl0f2:
        cmp byte ptr [esi+11],0
        jne short c4srl0c
        or cl,cl
        jz short c4srl0c
        mov ecx,[edi]
        cmp ecx,[esi]
        jne short c4srl0c
        mov ecx,[edi+4]
        cmp ecx,[esi+4]
        jne short c4srl0c
        mov cx,[edi+8]
        cmp cx,[esi+8]
        jne short c4srl0c
        jmp short c4srd
c4srl0f0:
        xor bl,bl
        mov bh,ah
c4srl0c:
        inc ah
        jmp c4srl0
c4srl0d:
        push offset c4srd
        or bl,bl
        jnz c4pu
        jmp c4rs
c4srf2:
        movsd
        movsd
        movsd
        sub esi,12
        jmp short c4srd
c4srf1:
        call c4rs
c4srd:
        mov word ptr [esi+10],0
        pop edi esi ecx ebx eax
        ret

;
; Examine right side expression of instruction
; In:
;   EBP -> current instruction
; Out:
;   CF=1 - expression value not found anywhere
;   CF=0 - expression value found somewhere
;   AL - expression type
;   AH - valid register bitmap for all regs except used in expression
;          bit 0: all needed values not in regs
;   BX - location of value of expression (if any) (BL=0f8h means not found)
;   CX - location of base (if any)
;   DX - location of index (if any)
;
c4rsex:
        push edi
        movzx eax,byte ptr [ebp+13]
        or al,al
        jz short c4rsexf0
        call c4ft
        mov dx,bx
        mov ah,1
        or bl,bl
        jz short c4rsexf0
        movzx edi,bl
        mov ah,c4regmask[edi]
c4rsexf0:
        mov al,[ebp+15]
        or al,al
        jz short c4rsexf1
        call c4ft
        mov cx,bx
        mov al,1
        or bl,bl
        jz short c4rsexf1
        movzx edi,bl
        or ah,c4regmask[edi]
        xor al,al
c4rsexf1:
        or ah,al
        lea edi,[ebp+12]
        call c4ce
        call c4fe
        jnc short c4rsexf2
        mov bl,0f8h
c4rsexf2:
        pop edi
        ret

;
; Load any registers needed in right side
; In:
;   AH - valid register bitmap for selection
;   CX - location of base (if any)
;   DX - location of index (if any)
;   EBP -> current instruction
; Out:
;   AH - valid register bitmap for all regs except loaded (CL,DL (if any))
;   CX - location of base (if any) (in reg)
;   DX - location of index (if any) (in reg)
;
c4rslr:
        push eax ebx
        mov byte ptr [esp+5],0
        mov al,[ebp+13]
        or al,al
        jz short c4rslrf0
        or dl,dl
        jnz short c4rslrf1
        call c4arbsr
        mov ebx,edx
        call c4rl
        mov dl,al
        cmp cx,bx
        jne short c4rslrf1
        mov cl,al
c4rslrf1:
        movzx ebx,dl
        mov al,c4regmask[ebx]
        or ah,al
        or [esp+5],al
c4rslrf0:
        mov al,[ebp+15]
        or al,al
        jz short c4rslrf2
        or cl,cl
        jnz short c4rslrf3
        call c4arbsr
        mov ebx,ecx
        call c4rl
        mov cl,al
c4rslrf3:
        movzx ebx,cl
        mov al,c4regmask[ebx]
        or [esp+5],al
c4rslrf2:
        pop ebx eax
        ret

;
; Load value of right side into register if not in reg already
; In:
;   AL - expression type
;   AH - valid register bitmap for selection
;   BX - location of value of expression (if any) (BL=0f8h means not found)
;   CX - location of base (if any) (must be reg)
;   DX - location of index (if any) (must be reg)
;   EBP -> current instruction
; Out:
;   AL - reg
;   AH - valid register bitmap for all regs except loaded (AL)
;
c4rslv:
        push esi edi
        lea edi,[ebp+12]
        cmp bl,0
        jle short c4rslvf0
        mov al,bl
        jmp c4rslvd
c4rslvf0:
        cmp al,1
        ja short c4rslvf2
        je short c4rslvf1
        mov al,cl
        jmp c4rslvd
c4rslvf1:
        call c4rslvr0
        call c4putexp
        @putstr crlf
        jmp short c4rslvf3
c4rslvf2:
        cmp byte ptr [edi],0
        je short c4rslvf4
        or bl,bl
        jnz short c4rslvf5
        call c4arbsr
        call c4rl
        jmp short c4rslvd
c4rslvf5:
        call c4rslvr0
        @putstr addxlds
        call c4putexp
        @putstr addxrcrlf
        jmp short c4rslvf3
c4rslvf4:
        call c4arbsr
        mov esi,eax
        @putop C4LEA
        mov eax,esi
        call c4putreg
        @putstr commaaddxl
        call c4putexp
        @putstr addxrcrlf
c4rslvf3:
        movzx esi,al
        lea esi,[esi*2+esi-3]
        lea esi,[esi*4+offset c4reg]
        mov edi,[ebp+12]
        mov [esi],edi
        mov edi,[ebp+16]
        mov [esi+4],edi
        movzx edi,word ptr [ebp+20]
        or edi,10000h
        mov [esi+8],edi
c4rslvd:
        movzx esi,al
        mov ah,c4regmask[esi]
        pop edi esi
        ret
;-----------------------------------------------------------------------------
c4rslvr0:
        call c4arbsr
        mov esi,eax
        @putop C1MOV
        mov eax,esi
        call c4putreg
        @putstr comma
        ret

