; Can be used multiple times per run, but memory must be freed and any fatal
;  error terminates immediately.

CRLFT           equ     0dh,0ah,'$'

;
; DATA
;
align 4
_ioibaseptr     dd      ?               ; base of input data
_ioiendptr      dd      ?               ; end of input data
_ioicurptr      dd      ?               ; current loc in input data
_ioobaseptr     dd      ?               ; base of output data area
_iooendptr      dd      ?               ; end of output data area
_ioocurptr      dd      ?               ; current loc in output data area

_iolexemeptr    dd      ?               ; ptr to last lexeme read
_iolexemelen    dd      ?               ; length of last lexeme read

line            dw      ?               ; current line number
defregspreserve db      ?               ; default bitmap of regs to preserve

errmsg0         db      'Error at _____: $'
errmsg1         db      'Fatal error: $'

ferr0           db      'Ran out of memory.',CRLFT
ferr1           db      'Could not open input file.',CRLFT
ferr2           db      'Could not create output file.',CRLFT

iorterr0        db      'Unexpected character.',CRLFT
iorterr1        db      'Unexpected end of line.',CRLFT

;
; CODE
;

;
; Put fatal error message and exit
; In:
;   EDX -> CRLFT terminated error message
;
_io_fatalerror:
        mov _exitcode,-1
        push edx
        mov edx,offset errmsg1
        call _putdosmsg
        pop edx
        call _putdosmsg
        jmp _exit

;
; Put error message with line number
; In:
;   EDX -> CRLFT terminated error message
;
_io_error:
        push eax edx
        mov _exitcode,-1
        mov ax,line
        call _cvhexw2bcd
        mov edx,offset errmsg0+9
        call putdecnum
        sub edx,9
        call _putdosmsg
        pop edx
        call _putdosmsg
        pop eax
        ret

;
; Init IO thingy, load input file and allocate output area
; In:
;   EDX -> input filename
;
_io_init:
        push eax ecx edx

        mov line,1
        mov _exitcode,0
        mov defregspreserve,0

        call _openfile
        mov edx,offset ferr1
        jc _io_fatalerror
        call _filesize
        mov ecx,eax
        inc eax
        call _getmem
        mov edx,offset ferr0
        jc _io_fatalerror
        mov edx,eax
        mov _ioibaseptr,eax
        mov _ioicurptr,eax
        mov _lexdbaseptr,eax
        add eax,ecx
        mov _ioiendptr,eax
        mov _lexdendptr,eax
        mov byte ptr [eax],0
        call _readfile
        call _closefile

        call _lomemsize
        mov ecx,eax
        call _himemsize
        mov edx,offset _gethimem
        cmp eax,ecx
        ja short ioinitf0
        mov eax,ecx
        mov edx,offset _getlomem
ioinitf0:
        mov ecx,eax
        call edx
        mov _ioobaseptr,eax
        mov _ioocurptr,eax
        add eax,ecx
        mov _iooendptr,eax

        pop edx ecx eax
        ret

;
; Finish IO stuff, write output file to disk
; In:
;   EDX -> output filename
;
_io_done:
        push eax ecx edx

        call _createfile
        mov edx,offset ferr2
        jc _io_fatalerror
        mov edx,_ioobaseptr
        mov ecx,_ioocurptr
        sub ecx,edx
        call _writefile
        call _closefile

        pop edx ecx eax
        ret

;
; Find next statement in input
; Out:
;   CF=0 - found statement
;   CF=1 - done with file
;
_io_find:
        push eax ebx ecx edi esi edi

        mov edi,_ioicurptr
        mov edx,offset ferr0
        mov ecx,_ioiendptr
        sub ecx,edi
        mov al,0dh
iofindl0:
        cmp ecx,1
        jc iofindf0
        cmp byte ptr [edi],'#'
        je short iofindl0d
iofindl1:
        repne scasb
        stc
        jne iofindf0
        cmp byte ptr [edi],0ah
        jne iofindl1
        sub ecx,1
        jc iofindf0
        inc edi
        inc line
        jmp iofindl0

iofindl0d:
        lea ebx,[edi+1]
        mov _iolexemeptr,ebx
        mov _iolexemelen,0

        mov esi,_ioicurptr
        mov ecx,edi
        sub ecx,esi
        mov edi,_ioocurptr
        lea ebx,[edi+ecx+1]
        cmp ebx,_iooendptr
        ja _io_fatalerror

        rep movsb
        mov byte ptr [edi],';'
        inc edi

        xchg esi,edi
        mov ecx,_ioiendptr
        sub ecx,edi
        mov ebx,edi
iofindl2:
        repne scasb
        jne short iofindl2d
        jecxz short iofindl2d
        cmp byte ptr [edi],0ah
        jne iofindl2
        dec edi
        inc ecx

iofindl2d:
        mov _ioicurptr,edi
        mov ecx,edi
        sub ecx,ebx
        mov edi,esi
        mov esi,ebx

        lea ebx,[edi+ecx+2]
        cmp ebx,_iooendptr
        ja _io_fatalerror
        rep movsb
        mov word ptr [edi],0a0dh
        add edi,2
        mov _ioocurptr,edi
        jmp short iofindf1

iofindf0:
        mov esi,_ioicurptr
        mov ecx,edi
        sub ecx,esi
        mov edi,_ioocurptr
        lea ebx,[edi+ecx]
        cmp ebx,_iooendptr
        ja _io_fatalerror
        rep movsb
        mov _ioocurptr,edi

iofindf1:
        pop edi esi edx ecx ebx eax
        ret

;
; Read next token from statement
; Out:
;   CF=0 - token gotten fine
;     EAX - token
;   CF=1 - error in input (abort this line)
;     EAX - ?
;
_io_read:
        push ecx edx
        mov edx,_iolexemeptr
        add edx,_iolexemelen
        mov _iolexemeptr,edx
        call _lex_findtoken
        mov _iolexemelen,ecx
        jnc short ioreadf0
        mov ecx,edx
        mov edx,offset iorterr1
        cmp ecx,_ioiendptr
        jae short ioreadf1
        cmp word ptr [ecx],0a0dh
        je short ioreadf1
        mov edx,offset iorterr0
ioreadf1:
        call _io_error
        stc
ioreadf0:
        pop edx ecx
        ret

